/****************************************************************************
  $Header: /afs/ifh.de/user/d/desiati/mydisk/mum/RCS/mumf2k.cc,v 1.2.2.1 2001/07/03 12:30:05 desiati Exp desiati $

  F2K shell for MUM muon propagator
  Miocinovic 20001031

 ----------------------------------------------------------------------------
  $Log: mumf2k.cc,v $
  Revision 1.2.2.1  2001/07/03 12:30:05  desiati
  Corrected muon loop configuration and, presumably, the muon track length
  determination.

  Revision 1.2.1.1  2001/07/02 16:44:18  desiati
  included active volume and muon propagation in three regions:
  1. from ice surface to active volume (enew)
  2. inside active volume (earray2)
  3. from active volume to desappearence (enew)
  still muon track length NON CORRECTLY DETERMINED !!

  Revision 1.2  2001/06/26 17:07:44  desiati
  Introduced 3 more options: Radius and height of active volume cilinder,
  and depth of amanda center detector
  Soubroutine enew for energy loss simulation outside active volume:
  defined in cfortran but not still implemented

  Revision 1.1  2001/06/26 13:40:26  desiati
  Initial revision

 ****************************************************************************/

#include <cstring>
#include <getopt.h>
#define f2cFortran
#include <cfortran.h>

#define MUMF2K_VERSION "0.2"
#include <offline_event++.h>    // Header needed for use of libevent++

#define SPEED_C   0.299792458   /* speed of light in vacuum (m/ns) */
#define EHIST_MAX  20000         /* EHistory array lengths, from mum.f */
#define VHIST_MAX  10000         /* VHistory array lengths, from mum.f */

#define Rad(x) ((x)*M_PI/180.)

/* Declaration of functions from mum.f */
PROTOCCALLSFSUB4(INIT_MU_,init_mu_,INT,INT,FLOAT,FLOAT)
#define INIT_MU(A,B,C,D) CCALLSFSUB4(INIT_MU_,init_mu_,INT,INT,FLOAT,FLOAT,A,B,C,D)

PROTOCCALLSFSUB2(EARRAY1,earray1,FLOAT,FLOAT)
#define EARRAY1(A,B) CCALLSFSUB2(EARRAY1,earray1,FLOAT,FLOAT,A,B)
PROTOCCALLSFFUN2(FLOAT,ENEW,enew,FLOAT,FLOAT)
#define ENEW(A,B) CCALLSFFUN2(ENEW,enew,FLOAT,FLOAT,A,B)

PROTOCCALLSFSUB5(EARRAY2,earray2,FLOAT,PDOUBLE,PFLOAT,PFLOAT,PINT)
#define EARRAY2(A,B,C,D,E) CCALLSFSUB5(EARRAY2,earray2,FLOAT,PDOUBLE,PFLOAT,PFLOAT,PINT,A,B,C,D,E)

/* History structure that let's us talk to mum common block ehistory,
   from earray1 :
   common /ehistory/ ener1_e,ener2_e,eleng_e,etr,ityp_e,numb_e */

typedef struct {
  double ener1_e[EHIST_MAX];
  double ener2_e[EHIST_MAX];
  float eleng_e[EHIST_MAX];
  float etr[EHIST_MAX];
  int ityp_e[EHIST_MAX];
  int numb_e;
} EHISTORY_DEF;
#define EHistory COMMON_BLOCK(EHISTORY, ehistory)
COMMON_BLOCK_DEF(EHISTORY_DEF, EHISTORY);
EHISTORY_DEF EHistory;

/* History structure that let's us talk to mum common block vhistory,
   from enew :
   common /vhistory/ numb,ityp,eleng,ener1,ener2 */

typedef struct {
  int numb;
  int ityp[VHIST_MAX];
  float eleng[VHIST_MAX];
  float ener1[VHIST_MAX];
  float ener2[VHIST_MAX];
} VHISTORY_DEF;
#define VHistory COMMON_BLOCK(VHISTORY, vhistory)
COMMON_BLOCK_DEF(VHISTORY_DEF, VHISTORY);
VHISTORY_DEF VHistory;

/* Local prototypes */
struct {
  int verbose;
  int n;
  int medium;
  int model;
  float em,vm;
  float range;
  char *infile;
  float radius;
  float height;
  float cdepth;
} opt = {0,0,3,1,0.5,1e-3,10000.,NULL,400.,800.,1740.};
int parse_input(int,char**);
void usage(char*);

int main(int argc, char *argv[]){
  int enr=0;  // Event counter 
  int gap=1;
  // Setup error handler
  errorHandler = new F2000Error(cerr,ErrorLevel::Errors);
  // Define data objects
  DataHeader<> header;
  DataEvent<> event;

  // Parse input
  if(parse_input(argc,argv)) usage(argv[0]);

  // Open I/O streams
  DataInterface input_file;
  if(opt.infile[0]!='\0')
    input_file.open(opt.infile);
  else
    input_file.open(cin);
  DataInterface output_file(cout);

  // Prepare muon propagator 
  // 3 == ice, 1 == interaction model, 0.5 == lower cut off (gev), 
  // 1e-3 == relative lower cut off
  INIT_MU(opt.medium,opt.model,opt.em,opt.vm);

  // Prepare F2000 I/O
  // Read in the header from an input stream
  header.read(input_file);
  // Record that this program modifed data
  header.file.historyList.push_back(HistoryInfo(argc,argv,MUMF2K_VERSION));
  // Write out the header
  header.write(output_file);

  // Loop over all events
  while(1){
    event.clear();
    if((enr<opt.n || !opt.n) && event.read(input_file)){ // Event succesfully read
      ++enr;
      vector<DataEvent<>::track_type> muTracks;
      int maxId=0;
      for(ioVector<DataEvent<>::track_type>::const_iterator track=
	    event.trackList.begin();track!=event.trackList.end();track++){
	if(strstr(track->getParticle().id().c_str(),"mu")) // It's muon
	  muTracks.push_back(*track);
	maxId=maxId>(int)track->getTag()?maxId:(int)track->getTag();
      }
      for(vector<DataEvent<>::track_type>::const_iterator 
	    track=muTracks.begin();track!=muTracks.end();track++)
      {							// Read Muons
	// Get parent's parameters (to avoid repeated lookup)
	float en_init = track->getEnergy();
	float zenith = track->getZenith();
	float azimuth = track->getAzimuth();
	float costh = cos(Rad(zenith)), sinth = sin(Rad(zenith));
	float cosph = cos(Rad(azimuth)), sinph = sin(Rad(azimuth));
	Value parent_tag = track->getTag();
	float x = track->getX();
	float y = track->getY();
	float z = track->getZ();
	float time = track->getTime();
	double dirx,diry,dirz; 

	// Is this muon track hitting the ACTIVE CYLINDER ?
	float z1, z2, h1, h2;
	float delta2, delta, beta, rho;
	beta = x*cosph+y*sinph;
	delta2 = beta*beta + opt.radius*opt.radius - (x*x+y*y);
	if (delta2 > 0)						// is the muon track intersecting the Cylinder ?
	{
	  delta = sqrt(delta2);
	  z1 = z - (beta - delta)*(costh/sinth);
	  z2 = z - (beta + delta)*(costh/sinth);
	  if ((z1 < -opt.height/2) || (z2 > opt.height/2))	// the muon track DOES NOT hit the cylinder
	  {
	    h1 = z; h2 = z;					// then does not propagate muon
	  }
	  else							// the muon HITS the cylinder sides
	  {
	    h1 = min(z1, opt.height/2); h2 = max(z2, -opt.height/2);
	  }
	}
	else							// delta2 < 0 : NO intersection
	{
	  h1 = z; h2 = z;
	}

	if (opt.verbose)
	  printf("costh:%f phi:%f x0:%f y0:%f z0:%f H1:%f H2:%f\n",
		 costh, azimuth, x, y, z, h1, h2);

	// Propagator variables 
	float en = en_init;
	double en_new;
	float etr;
	int jtyp;
	float depth_in = (z-h1)/costh;
	float depth_out = (z-h2)/costh;
	float path,tot_path=0.;

	// Calculate direction cosines (they are negative from muon angles)
	dirx = -cosph*sinth;
	diry = -sinph*sinth;
	dirz = -costh;

	// Quick estimate how many showers we'll produce
	int guess=(int)(pow(log10(en_init),3.7)*opt.range/2000.);
	event.trackList.reserve(maxId+guess);

	// Single step muon propagation from snow surface to active volume
	if (opt.verbose)
	  printf("ENEW: en=%f depth_in=%f depth_out=%f\n",en,depth_in,depth_out);

	en = ENEW(en, depth_in*100.);				// gives en after propagation at depth_in

	if (opt.verbose)
	  printf("ENEW(1): en_new=%f deltae=%f depth=%f\n", en, (en_init-en), VHistory.eleng[VHistory.numb-1]/100.);

	if (VHistory.ityp[VHistory.numb-1]==1)			// muon has reached depth_in => enters active volume
        {
	  tot_path += depth_in;					// tot_path from surface to active volume

	  // Loop while muon is alive
	  while(en>10.1 && tot_path<depth_out)			// minimum threshold = 10.1 GeV && inside active volume
	  {
	    EARRAY2(en,en_new,etr,path,jtyp);

	    tot_path += path/100.;  				// Keep in meters
	    en = en_new;
	    // Check return type and act accordingly
	    if(jtyp==2){ 					// Parent stopping point
	      vector<DataEvent<>::track_type>::iterator parent =  
	        event.trackList.find(parent_tag,
				   &DataEvent<>::track_type::getTag);
	      parent->setLength(tot_path); 			// muon stops inside active volume
	    }else if(jtyp>=4){ 				// Shower occurance
	      // Create and fill the shower
	      TrackInfo shower;
	      shower.setTag(++maxId);
	      shower.setParent(parent_tag);
	      switch(jtyp){
	      case 4 : shower.setParticle(ParticleType(BREMS_TOK)); break;
	      case 5 : shower.setParticle(ParticleType(EPAIR_TOK)); break;
	      case 6 : shower.setParticle(ParticleType(HADR_TOK)); break;
	      case 7 : shower.setParticle(ParticleType(DELTA_TOK)); break;
	      default: shower.setParticle(ParticleType(UNAVAILABLE_TOK)); break;
	      }
	      shower.setZenith(zenith);
	      shower.setAzimuth(azimuth);
	      shower.setX(x+dirx*tot_path);
	      shower.setY(y+diry*tot_path);
	      shower.setZ(z+dirz*tot_path);
	      shower.setTime(time+tot_path/SPEED_C);
	      shower.setEnergy(etr);
	      // Push shower onto track list
	      event.trackList.push_back(shower);
	    }
	  }							// End of Muon Loop while alive

	  if ((en>10.1) && (tot_path>=depth_out))		// muon with E>10.1 GeV has reached depth_out => exits active volume
	  {
	    en = ENEW(en, opt.range*100.);			// propagates muon until it stops
	    tot_path = depth_out + VHistory.eleng[VHistory.numb-1]/100.;

	    if (opt.verbose)
	      printf("ENEW(2): en=%f deltae=%f depth=%f\n", en, (en_init-en), depth_out+(VHistory.eleng[VHistory.numb-1])/100.);
	  }							// muon had reached depth_out => exited active volume
	}							// muon had reached depth_in => entered active volume
	else if (VHistory.eleng[VHistory.numb-1]<depth_in*100.)	// muon stops before reaching active volume
	  tot_path = VHistory.eleng[VHistory.numb-1]/100.;

	vector<DataEvent<>::track_type>::iterator parent =  
		event.trackList.find(parent_tag,&DataEvent<>::track_type::getTag);
	parent->setLength(tot_path);
      }								// End of Read Muons

      // Output event
      event.write(output_file);
      // Report progress
      if(opt.verbose)
	if(enr%gap==0){
	  cerr << "Processed " << enr << " events.\n";
	  gap*=2;
	}
    }else   // Read error or last event read 
      break;
  }
  
  // Read in the footer. 
  header.read(input_file);
  // Write out the footer
  header.write(output_file);

  // Report progress
  if(opt.verbose)
    cerr << "Processed total of " << enr 
	 << " events. Thank you for choosing Amasoft!\n";
 
  // Close data streams. Not explicitly necessary, but good style.
  input_file.close();
  output_file.close();
  exit(0);
}

#define OPTSTRING "hM:m:E:V:r:n:R:H:D:v"
int parse_input(int argc,char **argv){
  int c;                                                      /* option char */
  extern char *optarg;                              /* for use with getopt() */
  extern int optind, optopt;                        /* for use with getopt() */
  int errflg = 0;                               /* error flag (return value) */

  while ((c = getopt(argc, argv, OPTSTRING)) != -1)    /* search for options */
    switch (c) {
    case 'n': opt.n=atoi(optarg);break;
    case 'M': opt.medium = atoi(optarg); break;
    case 'm': opt.model = atoi(optarg); break;
    case 'E': opt.em = atof(optarg); break;
    case 'V': opt.vm = atof(optarg); break;
    case 'r': opt.range = atof(optarg); break;
    case 'R': opt.radius = atof(optarg); break;
    case 'H': opt.height = atof(optarg); break;
    case 'D': opt.cdepth = atof(optarg); break;
    case 'v': opt.verbose++; break;
    case 'h': errflg++; break;
    case ':':                                        /* -n without arguments */
      fprintf(stderr,"Option -%c requires an argument\n", optopt);
      errflg++;
      break;
    case '?':                                              /* invalid option */
      fprintf(stderr,"Unrecognized option: - %c\n", optopt);
      errflg++;
    } /* switch c */

  opt.infile = (optind >= argc)?(char *)"":argv[optind];

  return errflg; 
} /* parse_input() */
  
void usage(char *prog){
  cerr << prog << " Version " << MUMF2K_VERSION << "\n";
  cerr << "Usage: " << prog << "[arguments] file\n"
       << " arguments: -M n      [3=ice]  Medium n\n";
  cerr << "            -m model  [1]      Interaction model\n";
  cerr << "            -E emin   [0.5]    Lower interaction energy cut-off in GeV\n";
  cerr << "            -V vmin   [0.001]  Relative interaction energy cut-off\n";
  cerr << "            -r range  [10000]  Maximum tracking range x meters\n";
  cerr << "            -n n               process first n events only\n";
  cerr << "            -R radius [400]    Radius of Cilindric Active Volume\n";
  cerr << "            -H height [800]    Height of Cilindric Active Volume\n";
  cerr << "            -D depth  [1730]   Depth of AMANDA B10 center\n";
  cerr << "            -v                 verbose\n";
  cerr << "            -h                 this help\n";
  cerr << "            file               Input file with parent muon tracks,\n";
  cerr << "                      		if missing standard input is used\n";
  cerr << " For explanation of input parameter ranges, see mum.f code comments\n";
  exit(-1);
} /* usage() */
